==> untahris-2.0-src.tgz: untahris/castle.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Castle & Cannonball.

#include <math.h>

bool flags[128][128];

int qs, qe;
int qx[20000], qy[20000];

int bre;
int brx[20000], bry[20000];

struct OCannonball : TPlayObject {
  public:
  int fx, fy, tx, ty, stime;
  
  int getOx() {
    return Ctr(fx)+512;
    }

  int getOy() {
    return Ctr(fy)+512;
    }
  
  OCannonball(int _owner, int _fx, int _fy, int _tx, int _ty) : 
    fx(_fx), fy(_fy), tx(_tx), ty(_ty) {
    owner = _owner;
    stime = g.time;
    time = stime + (calcDistance(getOx()-tx, getOy()-ty)*spd()) / 500;
    g.objects.push_back(this);
    }
  OCannonball(TBuffer &b) {
    b.get(owner); b.get(time); b.get(stime);
    b.get(fx); b.get(fy); b.get(tx); b.get(ty); 
    }
  void save(TBuffer &b) {
    b.putChar(7);
    b.put(owner); b.put(time); b.put(stime);
    b.put(fx); b.put(fy); b.put(tx); b.put(ty); 
    }
  void go() {
    if(g.map[fy][fx] == cCannonA)
      g.mapSet(fx,fy, cCannonF, g.owner[fy][fx]);
    destroyAt(DeCtr(tx), DeCtr(ty), owner);
    g.deleteObject(this);
    }
  void draw() {
    int x = getOx();
    int y = getOy();
    x = x + ((tx-x) * (g.time-stime)) / (time-stime);
    y = y + ((ty-y) * (g.time-stime)) / (time-stime);
    DIc(icBall).drawAtCentered(CTG(x), CTG(y));
    }
  };

struct OCastle : TPlayObject {
  public:
  int bx, by; // brick coordinates
  int cx, cy; // cannon coordinates
  int tx, ty; // target coordinates
  int fx, fy; // cannon-to-fire coordinates
  bool brick[3][3];
  int utime;
  int mode;

  void ref1() {
    for(int x=0; x<3; x++) swap(brick[0][x], brick[2][x]);
    }
  
  void ref2() {
    for(int x=0; x<3; x++) for(int y=0; y<x; y++)
      swap(brick[y][x], brick[x][y]);
    }
  
  bool createNewBrick() {
    char* bricks[11] = {
      ".#..#..##",
      ".#..#..##",
      ".#..#.##.",
      ".#.##....",
      ".#.##....",
      ".#.##....",
      ".#..#..#.",
      ".#..#..#.",
      "##..#.##.",
      "##..#.##.",
      "....#...."};
    
    int r = g.rand() % 11;
    for(int y=0; y<3; y++) for(int x=0; x<3; x++)
      brick[y][x] = bricks[r][y*3+x] == '#';
    for(int u=g.rand() % 4; u<4; u++) ref1(), ref2();
    }
  
  OCastle(int _owner, int _x, int _y) {
    owner = _owner;
    time = g.time;
    utime= g.time;
    fx = 0;
    fy = 0;
    cx = _x;
    cy = _y;
    bx = _x-1;
    by = _y-1;
    tx = Ctr(_x);
    ty = Ctr(_y);
    mode = 0;
    createNewBrick();
    g.objects.push_back(this);
    }
  
  OCastle(TBuffer &b) {
    b.get(owner); b.get(time); b.get(mode);
    b.get(bx); b.get(by); b.get(cx); b.get(cy); b.get(tx); b.get(ty);
    b.get(fx); b.get(fy); b.get(utime);
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) b.get(brick[ky][kx]);
    }
  void save(TBuffer &b) {
    b.putChar(6);
    b.put(owner); b.put(time); b.put(mode);
    b.put(bx); b.put(by); b.put(cx); b.put(cy); b.put(tx); b.put(ty);
    b.put(fx); b.put(fy); b.put(utime);
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) b.put(brick[ky][kx]);
    }

  void checkEnclosed(int rx, int ry) {
    if(!isCastlable(g.map[ry][rx])) return;
    if(g.owner[ry][rx] == owner) return; // already ours...
    qs = 0;
    qe = 1;
    qx[0] = rx; qy[0] = ry;
    bool ok = true;
    while(qs<qe) {
      int rx = qx[qs];
      int ry = qy[qs];
      qs++;
      if(g.map[ry][rx] == cPermaWall) {ok = false; break;}
      if(isCastlable(g.map[ry][rx])) {
        for(int lx=rx-1; lx<=rx+1; lx++)
        for(int ly=ry-1; ly<=ry+1; ly++) {
          if(flags[ly][lx]) continue;
          flags[ly][lx] = true;
          qx[qe] = lx; qy[qe] = ly;
          qe++;
          }
        }
      }
    if(ok) {
      for(int q=0; q<qe; q++) {
        rx = qx[q]; ry = qy[q];
        if(isCastlable(g.map[ry][rx])) g.mapSet(rx,ry, g.map[ry][rx], owner);
        flags[ry][rx] = false;
        }
      }
    else for(int q=0; q<qe; q++) flags[qy[q]][qx[q]] = false;
    }

  void placeBrick() {
    bool ok = false;
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) if(brick[ky][kx]) {
      if(g.map[by+ky][bx+kx] != cFree) return;
      for(int ly=-2; ly<3; ly++)
      for(int lx=-2; lx<3; lx++)
      if(by+ky+ly >= 0 && by+ky+ly < g.sizeY &&
        bx+kx+lx >= 0 && bx+kx+lx < g.sizeX &&
        g.owner[by+ky+ly][bx+kx+lx] == owner
        )
        ok = true;
      }
    if(ok) {
      d.playSoundAt(d.sndPlace, bx+1, by+1);
      for(int ky=0; ky<3; ky++)
      for(int kx=0; kx<3; kx++) if(brick[ky][kx]) {
        g.mapSet(bx+kx, by+ky, cCastleWall, owner);
        for(int tx=0; tx<3; tx++)
        for(int ty=0; ty<3; ty++)
          checkEnclosed(bx+kx+tx-1, by+ky+ty-1);
        }
      }
    createNewBrick();
    }
  
  void placeCannon() {
    for(int ky=0; ky<2; ky++)
    for(int kx=0; kx<2; kx++)
      if(g.map[cy+ky][cx+kx] != cFree || g.owner[cy+ky][cx+kx] != owner)
        return;
    if(ow()->useAmmo(50)) {
      d.playSoundAt(d.sndPlace, fx, fy);
      placeBig(cx, cy, cCannonF, owner);
      }
    }
    
  void fireCannon() {
    if(fx == 999 || fy == 999) return;
    if(g.map[fy][fx] != cCannonF) return;
    if(g.owner[fy][fx] != owner) return;
    g.mapSet(fx,fy, cCannonA, owner);
    d.playSoundAt(d.sndShoot, fx, fy);
    new OCannonball(owner, fx, fy, tx, ty);
    }
  
  void reactToKey(PlayerKey pk, bool pressed) {
    if(!pressed) return;
    if(pk == keyEarth) {
      mode++;
      mode %= 3;
      }
    else switch(mode) {
      case 0:
        switch(pk) {
          case keyLeft:
            if(bx>0) bx--;
            break;
          case keyRight:
            if(bx<g.sizeX-3) bx++;
            break;
          case keyDown:
            if(by<g.sizeY-3) by++;
            break;
          case keyUp:
            if(by>0) by--;
            break;
          case keyFire: 
            placeBrick();
            break;
          case keyAir:
            ref1(); ref2();
            break;
          case keyWater:
            ref2(); ref1();
            break;
          }
        break;
      case 1:
        switch(pk) {
          case keyLeft:
            if(cx>1) cx--;
            break;
          case keyRight:
            if(cx<g.sizeX-2) cx++;
            break;
          case keyDown:
            if(cy<g.sizeY-2) cy++;
            break;
          case keyUp:
            if(cy>1) cy--;
            break;
          case keyFire: 
          case keyAir:
          case keyWater:
            placeCannon();
            break;
          }
        break;
      case 2:
        switch(pk) {
          case keyFire: 
            fireCannon();
            break;
          }
        break;
      }
    }
  
  int cannonDistance(int x, int y) {
    return calcDistance(tx-Ctr(x)-512, ty-Ctr(y)-512);
    }
  
  void manageCastle() {
    int levels=0, scores=0, ammos=0;
    bre = 0;
    fx = 999, fy = 999;
    for(int y=1; y<g.sizeY-1; y++)
    for(int x=1; x<g.sizeX-1; x++)
      if(g.owner[y][x] == owner && isCastlable(g.map[y][x])) {
        if(g.map[y][x] == cBonusLevel) levels++;
        if(g.map[y][x] == cBonusScore) scores++;
        if(g.map[y][x] == cBonusAmmo)  ammos++;
        if(g.map[y][x] == cCannonF) {
          if(cannonDistance(x,y) < cannonDistance(fx, fy))
            fx=x, fy=y;
          }
        for(int ay=y-1; ay<=y+1; ay++)
        for(int ax=x-1; ax<=x+1; ax++) {
          if(isCastlable(g.map[ay][ax]) && g.owner[ay][ax] == 0) {
            // maybe everything is OK...
            checkEnclosed(ax, ay);
            // ... or maybe not
            if(g.owner[ay][ax] == 0) {
              brx[bre] = x; bry[bre] = y; bre++;
              }
            }
          }
        }
    for(int t=0; t<bre; t++) {
      int ax = brx[t], ay = bry[t];
      g.mapSet(ax, ay, g.map[ay][ax], 0);                     
      }
    ow()->level = 100 * levels;
    int32& tv (ow()->typeval);
    if(scores == 0) {
      tv -= 10;
      if(tv <= 0) {
        // dies without sound
        killedBy(0);
        ow()->ammo = 10;
        return;
        }
      }
    if(tv < 500 * scores) {
      tv += 25;
      if(tv > 500 * scores) tv = 500 * scores;
      }
    ow()->ammo += ammos;
    }

  void go() {
    if(time >= utime) { manageCastle(); utime = time + 200; }
    if(mode == 2) {
      int spd = 128;
      if(ow()->key[keyAir]) spd *= 2;
      if(ow()->key[keyWater]) spd *= 2;
      if(ow()->key[keyRight] && tx < g.sizeX*1024 - 2048)
        tx += spd;
      if(ow()->key[keyDown] && ty < g.sizeY*1024 - 2048)
        ty += spd;
      if(ow()->key[keyLeft] && tx >= 2048)
        tx -= spd;
      if(ow()->key[keyUp] && ty >= 2048)
        ty -= spd;
      }
    time += 10;
    }

  void draw() {
    switch(mode) {
      case 0:
        for(int y=0; y<3; y++)
        for(int x=0; x<3; x++)
        if(brick[y][x]) {
          DIc(icCastle).drawAtCentered(
            d.ofsX + d.Res * (x+bx) + d.Res/2, 
            d.ofsY + d.Res * (y+by) + d.Res/2
            );
          }
        break;
      case 1:
        DIc(icCannonP).drawAtCentered(
          d.ofsX + d.Res * cx + d.Res, 
          d.ofsY + d.Res * cy + d.Res
          );
        break;
      case 2:
        DIc(icTarget).drawAtCentered(CTG(tx), CTG(ty));
        break;
      }
    }
  };

==> untahris-3.1-src.tgz: untahris/castle.cpp <==
// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Castle & Cannonball.

#include <math.h>

bool flags[128][128];

int qs, qe;
int qx[20000], qy[20000];

int bre;
int brx[20000], bry[20000];

struct OCannonball : TPlayObject {
  public:
  int32 fx, fy, tx, ty, stime;
  
  int getOx() {
    return Ctr(fx)+512;
    }

  int getOy() {
    return Ctr(fy)+512;
    }
  
  OCannonball(int _owner, int _fx, int _fy, int _tx, int _ty) : 
    fx(_fx), fy(_fy), tx(_tx), ty(_ty) {
    owner = _owner;
    stime = g.time;
    time = stime + (calcDistance(getOx()-tx, getOy()-ty)*spd()) / 500;
    g.objects.push_back(this);
    }
  OCannonball(TBuffer &b) {
    b.get(owner); b.get(time); b.get(stime);
    b.get(fx); b.get(fy); b.get(tx); b.get(ty); 
    }
  void save(TBuffer &b) {
    b.putChar(7);
    b.put(owner); b.put(time); b.put(stime);
    b.put(fx); b.put(fy); b.put(tx); b.put(ty); 
    }
  void go() {
    if(g.map[fy][fx] == cCannonA)
      g.mapSet(fx,fy, cCannonF, g.owner[fy][fx]);

    for(int o=0; o<size(g.objects); o++) if(g.objects[o])
      g.objects[o]->reactToCannon(this);

    destroyAt(DeCtr(tx), DeCtr(ty), owner);
    g.deleteObject(this);
    }
  void draw() {
    int x = getOx();
    int y = getOy();
    x = x + ((tx-x) * (g.time-stime)) / (time-stime);
    y = y + ((ty-y) * (g.time-stime)) / (time-stime);
    DIc(icBall).drawAtCentered(CTG(x), CTG(y));
    }
  };

struct OCastle : TPlayObject {
  public:
  int32 bx, by; // brick coordinates
  int32 cx, cy; // cannon coordinates
  int32 tx, ty; // target coordinates
  int32 fx, fy; // cannon-to-fire coordinates
  bool brick[3][3];
  int32 utime;
  int32 mode;

  void ref1() {
    for(int x=0; x<3; x++) swap(brick[0][x], brick[2][x]);
    }
  
  void ref2() {
    for(int x=0; x<3; x++) for(int y=0; y<x; y++)
      swap(brick[y][x], brick[x][y]);
    }
  
  bool createNewBrick() {
    char* bricks[11] = {
      ".#..#..##",
      ".#..#..##",
      ".#..#.##.",
      ".#.##....",
      ".#.##....",
      ".#.##....",
      ".#..#..#.",
      ".#..#..#.",
      "##..#.##.",
      "##..#.##.",
      "....#...."};
    
    int r = g.rand() % 11;
    for(int y=0; y<3; y++) for(int x=0; x<3; x++)
      brick[y][x] = bricks[r][y*3+x] == '#';
    for(int u=g.rand() % 4; u<4; u++) ref1(), ref2();
    }
  
  OCastle(int _owner, int _x, int _y) {
    owner = _owner;
    time = g.time;
    utime= g.time;
    fx = 0;
    fy = 0;
    cx = _x;
    cy = _y;
    bx = _x-1;
    by = _y-1;
    tx = Ctr(_x);
    ty = Ctr(_y);
    mode = 0;
    createNewBrick();
    g.objects.push_back(this);
    }
  
  OCastle(TBuffer &b) {
    b.get(owner); b.get(time); b.get(mode);
    b.get(bx); b.get(by); b.get(cx); b.get(cy); b.get(tx); b.get(ty);
    b.get(fx); b.get(fy); b.get(utime);
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) b.get(brick[ky][kx]);
    }
  void save(TBuffer &b) {
    b.putChar(6);
    b.put(owner); b.put(time); b.put(mode);
    b.put(bx); b.put(by); b.put(cx); b.put(cy); b.put(tx); b.put(ty);
    b.put(fx); b.put(fy); b.put(utime);
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) b.put(brick[ky][kx]);
    }

  void checkEnclosed(int rx, int ry) {
    if(!isCastlable(g.map[ry][rx])) return;
    if(g.owner[ry][rx] == owner) return; // already ours...
    qs = 0;
    qe = 1;
    qx[0] = rx; qy[0] = ry;
    bool ok = true;
    while(qs<qe) {
      int rx = qx[qs];
      int ry = qy[qs];
      qs++;
      if(g.map[ry][rx] == cPermaWall) {ok = false; break;}
      if(isCastlable(g.map[ry][rx])) {
        for(int lx=rx-1; lx<=rx+1; lx++)
        for(int ly=ry-1; ly<=ry+1; ly++) {
          if(flags[ly][lx]) continue;
          flags[ly][lx] = true;
          qx[qe] = lx; qy[qe] = ly;
          qe++;
          }
        }
      }
    if(ok) {
      for(int q=0; q<qe; q++) {
        rx = qx[q]; ry = qy[q];
        if(isCastlable(g.map[ry][rx])) g.mapSet(rx,ry, g.map[ry][rx], owner);
        flags[ry][rx] = false;
        }
      }
    else for(int q=0; q<qe; q++) flags[qy[q]][qx[q]] = false;
    }

  void placeBrick() {
    bool ok = false;
    for(int ky=0; ky<3; ky++)
    for(int kx=0; kx<3; kx++) if(brick[ky][kx]) {
      if(g.map[by+ky][bx+kx] != cFree) return;
      for(int ly=-2; ly<3; ly++)
      for(int lx=-2; lx<3; lx++)
      if(by+ky+ly >= 0 && by+ky+ly < g.sizeY &&
        bx+kx+lx >= 0 && bx+kx+lx < g.sizeX &&
        g.owner[by+ky+ly][bx+kx+lx] == owner
        )
        ok = true;
      }
    if(ok) {
      d.playSoundAt(d.sndPlace, bx+1, by+1);
      for(int ky=0; ky<3; ky++)
      for(int kx=0; kx<3; kx++) if(brick[ky][kx]) {
        g.mapSet(bx+kx, by+ky, cCastleWall, owner);
        for(int tx=0; tx<3; tx++)
        for(int ty=0; ty<3; ty++)
          checkEnclosed(bx+kx+tx-1, by+ky+ty-1);
        }
      }
    createNewBrick();
    }
  
  void placeCannon() {
    for(int ky=0; ky<2; ky++)
    for(int kx=0; kx<2; kx++)
      if(g.map[cy+ky][cx+kx] != cFree || g.owner[cy+ky][cx+kx] != owner)
        return;
    if(ow()->useAmmo(gc.ammoPerCannon)) {
      d.playSoundAt(d.sndPlace, fx, fy);
      placeBig(cx, cy, cCannonF, owner);
      }
    }
    
  void fireCannon() {
    if(fx == 999 || fy == 999) return;
    if(g.map[fy][fx] != cCannonF) return;
    if(g.owner[fy][fx] != owner) return;
    if(ow()->useAmmo()) {
      g.mapSet(fx,fy, cCannonA, owner);
      d.playSoundAt(d.sndShoot, fx, fy);
      new OCannonball(owner, fx, fy, tx, ty);
      }
    }
  
  void reactToKey(PlayerKey pk, bool pressed) {
    if(!pressed) return;
    if(pk == keyEarth) {
      mode++;
      mode %= 3;
      }
    else switch(mode) {
      case 0:
        switch(pk) {
          case keyLeft:
            if(bx>0) bx--;
            break;
          case keyRight:
            if(bx<g.sizeX-3) bx++;
            break;
          case keyDown:
            if(by<g.sizeY-3) by++;
            break;
          case keyUp:
            if(by>0) by--;
            break;
          case keyFire: 
            placeBrick();
            break;
          case keyAir:
            ref1(); ref2();
            break;
          case keyWater:
            ref2(); ref1();
            break;
          }
        break;
      case 1:
        switch(pk) {
          case keyLeft:
            if(cx>1) cx--;
            break;
          case keyRight:
            if(cx<g.sizeX-2) cx++;
            break;
          case keyDown:
            if(cy<g.sizeY-2) cy++;
            break;
          case keyUp:
            if(cy>1) cy--;
            break;
          case keyFire: 
          case keyAir:
          case keyWater:
            placeCannon();
            break;
          }
        break;
      case 2:
        switch(pk) {
          case keyFire: 
            fireCannon();
            break;
          }
        break;
      }
    }
  
  int cannonDistance(int x, int y) {
    return calcDistance(tx-Ctr(x)-512, ty-Ctr(y)-512);
    }
  
  void manageCastle() {
    int levels=0, scores=0, ammos=0;
    bre = 0;
    fx = 999, fy = 999;
    for(int y=1; y<g.sizeY-1; y++)
    for(int x=1; x<g.sizeX-1; x++)
      if(g.owner[y][x] == owner && isCastlable(g.map[y][x])) {
        if(g.map[y][x] == cBonusLevel) levels++;
        if(g.map[y][x] == cBonusScore) scores++;
        if(g.map[y][x] == cBonusAmmo)  ammos++;
        if(g.map[y][x] == cCannonF) {
          if(cannonDistance(x,y) < cannonDistance(fx, fy))
            fx=x, fy=y;
          }
        for(int ay=y-1; ay<=y+1; ay++)
        for(int ax=x-1; ax<=x+1; ax++) {
          if(isCastlable(g.map[ay][ax]) && g.owner[ay][ax] == 0) {
            // maybe everything is OK...
            checkEnclosed(ax, ay);
            // ... or maybe not
            if(g.owner[ay][ax] == 0) {
              brx[bre] = x; bry[bre] = y; bre++;
              }
            }
          }
        }
    for(int t=0; t<bre; t++) {
      int ax = brx[t], ay = bry[t];
      g.mapSet(ax, ay, g.map[ay][ax], 0);                     
      }
    ow()->level = levels;
    int32& tv (ow()->typeval);
    if(scores == 0) {
      tv -= gc.castleRotSpeed;
      if(tv <= 0) {
        // dies without sound
        killedBy(0);
        ow()->ammo = 0;
        return;
        }
      }
    scores *= gc.castlePowerBase;
    if(tv < scores) {
      tv += gc.castleRegenSpeed;
      if(tv > scores) tv = scores;
      }
    ow()->ammo += ammos * ec().bonusAmmo;
    }

  void go() {
    if(time >= utime) { manageCastle(); utime = time + 200; }
    if(mode == 2) {
      int spd = 128;
      if(ow()->key[keyAir]) spd *= 2;
      if(ow()->key[keyWater]) spd *= 2;
      if(ow()->key[keyRight] && tx < g.sizeX*1024 - 2048)
        tx += spd;
      if(ow()->key[keyDown] && ty < g.sizeY*1024 - 2048)
        ty += spd;
      if(ow()->key[keyLeft] && tx >= 2048)
        tx -= spd;
      if(ow()->key[keyUp] && ty >= 2048)
        ty -= spd;
      }
    time += 10;
    }

  void draw() {
    switch(mode) {
      case 0:
        for(int y=0; y<3; y++)
        for(int x=0; x<3; x++)
        if(brick[y][x]) {
          DIc(icCastle).drawAtCentered(
            d.Res * (x+bx) + d.Res/2, 
            d.Res * (y+by) + d.Res/2
            );
          }
        break;
      case 1:
        DIc(icCannonP).drawAtCentered(
          d.Res * cx + d.Res, 
          d.Res * cy + d.Res
          );
        break;
      case 2:
        DIc(icTarget).drawAtCentered(CTG(tx), CTG(ty));
        break;
      }
    }
  };
